/************************************************************************
 * Erdos-Renyi random graph generator
 * Copyright (C) 2011, Georgia Institute of Technology,
 * Author: Henning Meyerhenke
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>
 ***********************************************************************/

#include <cstdlib>
#include <cstdint>
#include <cstdio>
#include <cstring>
#include <string>
#include <cassert>
#include <vector>
#include <cmath>

using namespace std;

/**
 * Returns number of steps you need to wait until the next success (edge) occurs.
 */
static inline uint32_t get_next_edge_distance(const double log_cp) {
	return (uint32_t) 1 + floor(log(1.0 - drand48()) / log_cp);
}

/**
 * Creates Erdos-Renyi random graph G(n,prob) and writes it to a file
 * in format used for the 10th DIMACS Implementation Challenge.
 * The generation follows mostly Vladimir Batagelj and Ulrik Brandes: "Efficient
 * generation of large random networks", Phys Rev E 71, 036113 (2005).
 */
static
void graph_create_erdos_renyi_gnp(uint32_t n, const double prob,
		const string& filename) {
	// declare and init variables
	uint32_t i = 0;
	uint32_t j = -2;
	uint32_t step = 0;
	uint32_t num2E = 0;
	uint32_t estimated_num2E;
	const double log_cp = log(1.0 - prob); // log of counter probability

	// stores lower diagonal entries of each row before writing them to file
	vector<uint32_t>* row_entries = new vector<uint32_t> [n];

	// set all degrees to 0
	uint32_t* degree = new uint32_t[n];
	for (i = 0; i < n; ++i) {
		degree[i] = 0;
	}

	// create file
	FILE* file = fopen(filename.c_str(), "w");
	if (file == NULL) {
		printf("Error: File %s could not be created, abort.\n",
				filename.c_str());
		abort();
	}

	// print estimate because we don't know the number of edges yet
	estimated_num2E = (uint32_t) ((prob * (double) n) * (double) (n-1));
	fprintf(file, "%u %u \n", n, estimated_num2E / 2);

	// create edges
	i = 0;
	while (i < n) {
		// compute new step length
		step = get_next_edge_distance(log_cp);
		j += step;

		// ensure that we get no singleton node (NOTE: not actual part of ER model)
		if (degree[i] == 0 && j >= n) {
			j -= step;
			continue;
		}

		// check if at end of row
		while (j >= n && i < n) {
			// adapt to next row
			i++;
			j = (j - n) + 1 + i;

			if (i < n) {
				fprintf(file, "\n");

				// print row i up to diagonal
				for (uint32_t k = 0; k < row_entries[i].size(); ++k) {
					fprintf(file, "%u ", row_entries[i][k]);
				}
				fflush(file);
				row_entries[i].clear();
			}

#ifdef DEBUG
			if (i % 1048576 == 0)
				printf("Finished row %i.\n", i-1);
#endif
		}

		// insert edge
		if (i < n) {
			fprintf(file, "%u ", j + 1);
			row_entries[j].push_back(i + 1);

			++degree[i];
			++degree[j];
			num2E += 2;
		}
	}
	delete[] degree;

	// replace estimate by real number of edges (counted twice)
	rewind(file);
	fprintf(file, "%u %u", n, num2E / 2);

	delete[] row_entries;
	fclose(file);
#ifndef NDEBUG
	printf("Wrote graph file with %u vertices and %u undirected edges, difference to estimate: %i.\n", n, num2E / 2, (int) (estimated_num2E - (int) num2E) / 2);
#endif
}

static inline
void print_usage(char* argv[]) {
	printf("Erdos-Renyi random graph generator,\n" \
			"Copyright (C) 2011, Georgia Institute of Technology.\n\n" \
			"This program comes with ABSOLUTELY NO WARRANTY.\n" \
			"This is free software, and you are welcome to redistribute it " \
			"under certain conditions.\n");
	printf("\nUsage: %s <n> <factor for prob = factor * log n / n> <filename> [seed]\n",
			argv[0]);
}

int main(int argc, char* argv[]) {
	// check cmd line args
	if (argc <= 3) {
		print_usage(argv);
		abort();
	}

	// "parse" cmd line args
	uint32_t n = atoi(argv[1]);
	string filename(argv[3]);
	if (argc > 4) {
		srand48(atoi(argv[4]));
	}

	// call generator and print graph
	double prob = (double) (atof(argv[2]) * log(n) / (double) n);
	assert(n >= 16 && atof(argv[2]) >= 1.0);
	graph_create_erdos_renyi_gnp(n, prob, filename);

	return EXIT_SUCCESS;
}
